#!/usr/bin/python2.5

# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Narcissus JavaScript engine, written in Javascript.
#
# The Initial Developer of the Original Code is
# Brendan Eich <brendan@mozilla.org>.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# The Python version of the code was created by JT Olds <jtolds@xnet5.com>,
# and is a direct translation from the Javascript version.
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK ***** */

"""
 PyNarcissus

 A lexical scanner and parser. JS implemented in JS, ported to Python.
"""

__author__ = "JT Olds"
__author_email__ = "jtolds@xnet5.com"
__date__ = "2009-03-24"
__all__ = ["ParseError", "parse", "tokens"]

import re, sys, types

class Object: pass
class Error_(Exception): pass
class ParseError(Error_): pass

tokens = dict(enumerate((
        # End of source.
        "END",

        # Operators and punctuators. Some pair-wise order matters, e.g. (+, -)
        # and (UNARY_PLUS, UNARY_MINUS).
        "\n", ";",
        ",",
        "=",
        "?", ":", "CONDITIONAL",
        "||",
        "&&",
        "|",
        "^",
        "&",
        "==", "!=", "===", "!==",
        "<", "<=", ">=", ">",
        "<<", ">>", ">>>",
        "+", "-",
        "*", "/", "%",
        "!", "~", "UNARY_PLUS", "UNARY_MINUS",
        "++", "--",
        ".",
        "[", "]",
        "{", "}",
        "(", ")",

        # Nonterminal tree node type codes.
        "SCRIPT", "BLOCK", "LABEL", "FOR_IN", "CALL", "NEW_WITH_ARGS", "INDEX",
        "ARRAY_INIT", "OBJECT_INIT", "PROPERTY_INIT", "GETTER", "SETTER",
        "GROUP", "LIST",

        # Terminals.
        "IDENTIFIER", "NUMBER", "STRING", "REGEXP",

        # Keywords.
        "break",
        "case", "catch", "const", "continue",
        "debugger", "default", "delete", "do",
        "else", "enum",
        "false", "finally", "for", "function",
        "if", "in", "instanceof",
        "new", "null",
        "return",
        "switch",
        "this", "throw", "true", "try", "typeof",
        "var", "void",
        "while", "with")))

# Operator and punctuator mapping from token to tree node type name.
# NB: superstring tokens (e.g., ++) must come before their substring token
# counterparts (+ in the example), so that the opRegExp regular expression
# synthesized from this list makes the longest possible match.
opTypeNames = [
        ('\n',   "NEWLINE"),
        (';',    "SEMICOLON"),
        (',',    "COMMA"),
        ('?',    "HOOK"),
        (':',    "COLON"),
        ('||',   "OR"),
        ('&&',   "AND"),
        ('|',    "BITWISE_OR"),
        ('^',    "BITWISE_XOR"),
        ('&',    "BITWISE_AND"),
        ('===',  "STRICT_EQ"),
        ('==',   "EQ"),
        ('=',    "ASSIGN"),
        ('!==',  "STRICT_NE"),
        ('!=',   "NE"),
        ('<<',   "LSH"),
        ('<=',   "LE"),
        ('<',    "LT"),
        ('>>>',  "URSH"),
        ('>>',   "RSH"),
        ('>=',   "GE"),
        ('>',    "GT"),
        ('++',   "INCREMENT"),
        ('--',   "DECREMENT"),
        ('+',    "PLUS"),
        ('-',    "MINUS"),
        ('*',    "MUL"),
        ('/',    "DIV"),
        ('%',    "MOD"),
        ('!',    "NOT"),
        ('~',    "BITWISE_NOT"),
        ('.',    "DOT"),
        ('[',    "LEFT_BRACKET"),
        (']',    "RIGHT_BRACKET"),
        ('{',    "LEFT_CURLY"),
        ('}',    "RIGHT_CURLY"),
        ('(',    "LEFT_PAREN"),
        (')',    "RIGHT_PAREN"),
    ]

keywords = {}

# Define const END, etc., based on the token names.  Also map name to index.
for i, t in tokens.copy().iteritems():
    if re.match(r'^[a-z]', t):
        const_name = t.upper()
        keywords[t] = i
    elif re.match(r'^\W', t):
        const_name = dict(opTypeNames)[t]
    else:
        const_name = t
    globals()[const_name] = i
    tokens[t] = i

assignOps = {}

# Map assignment operators to their indexes in the tokens array.
for i, t in enumerate(['|', '^', '&', '<<', '>>', '>>>', '+', '-', '*', '/', '%']):
    assignOps[t] = tokens[t]
    assignOps[i] = t

# Build a regexp that recognizes operators and punctuators (except newline).
opRegExpSrc = "^"
for i, j in opTypeNames:
    if i == "\n": continue
    if opRegExpSrc != "^": opRegExpSrc += "|^"
    opRegExpSrc += re.sub(r'[?|^&(){}\[\]+\-*\/\.]', lambda x: "\\%s" % x.group(0), i)
opRegExp = re.compile(opRegExpSrc)

# Convert opTypeNames to an actual dictionary now that we don't care about ordering
opTypeNames = dict(opTypeNames)

# A regexp to match floating point literals (but not integer literals).
fpRegExp = re.compile(r'^\d+\.\d*(?:[eE][-+]?\d+)?|^\d+(?:\.\d*)?[eE][-+]?\d+|^\.\d+(?:[eE][-+]?\d+)?')

# A regexp to match regexp literals.
reRegExp = re.compile(r'^\/((?:\\.|\[(?:\\.|[^\]])*\]|[^\/])+)\/([gimy]*)')

class SyntaxError_(ParseError):
    def __init__(self, message, filename, lineno):
        ParseError.__init__(self, "Syntax error: %s\n%s:%s" %
                (message, filename, lineno))

class Tokenizer(object):
    def __init__(self, s, f, l):
        self.cursor = 0
        self.source = str(s)
        self.tokens = {}
        self.tokenIndex = 0
        self.lookahead = 0
        self.scanNewlines = False
        self.scanOperand = True
        self.filename = f
        self.lineno = l

    input_ = property(lambda self: self.source[self.cursor:])
    done = property(lambda self: self.peek() == END)
    token = property(lambda self: self.tokens.get(self.tokenIndex))

    def match(self, tt):
        return self.get() == tt or self.unget()

    def mustMatch(self, tt):
        if not self.match(tt):
            raise self.newSyntaxError("Missing " + tokens.get(tt).lower())
        return self.token

    def peek(self):
        if self.lookahead:
            next = self.tokens.get((self.tokenIndex + self.lookahead) & 3)
            if self.scanNewlines and (getattr(next, "lineno", None) !=
                    getattr(self, "lineno", None)):
                tt = NEWLINE
            else:
                tt = getattr(next, "type_", None)
        else:
            tt = self.get()
            self.unget()
        return tt

    def peekOnSameLine(self):
        self.scanNewlines = True
        tt = self.peek()
        self.scanNewlines = False
        return tt

    def get(self):
        while self.lookahead:
            self.lookahead -= 1
            self.tokenIndex = (self.tokenIndex + 1) & 3
            token = self.tokens.get(self.tokenIndex)
            if getattr(token, "type_", None) != NEWLINE or self.scanNewlines:
                return getattr(token, "type_", None)

        while True:
            input__ = self.input_
            if self.scanNewlines:
                match = re.match(r'^[ \t]+', input__)
            else:
                match = re.match(r'^\s+', input__)
            if match:
                spaces = match.group(0)
                self.cursor += len(spaces)
                newlines = re.findall(r'\n', spaces)
                if newlines:
                    self.lineno += len(newlines)
                input__ = self.input_

            match = re.match(r'^\/(?:\*(?:.|\n)*?\*\/|\/.*)', input__)
            if not match:
                break
            comment = match.group(0)
            self.cursor += len(comment)
            newlines = re.findall(r'\n', comment)
            if newlines:
                self.lineno += len(newlines)

        self.tokenIndex = (self.tokenIndex + 1) & 3
        token = self.tokens.get(self.tokenIndex)
        if not token:
            token = Object()
            self.tokens[self.tokenIndex] = token

        if not input__:
            token.type_ = END
            return END

        def matchInput():
            match = fpRegExp.match(input__)
            if match:
                token.type_ = NUMBER
                token.value = float(match.group(0))
                return match.group(0)

            match = re.match(r'^0[xX][\da-fA-F]+|^0[0-7]*|^\d+', input__)
            if match:
                token.type_ = NUMBER
                token.value = eval(match.group(0))
                return match.group(0)

            match = re.match(r'^[$_\w]+', input__)       # FIXME no ES3 unicode
            if match:
                id_ = match.group(0)
                token.type_ = keywords.get(id_, IDENTIFIER)
                token.value = id_
                return match.group(0)

            match = re.match(r'^"(?:\\.|[^"])*"|^\'(?:\\.|[^\'])*\'', input__)
            if match:
                token.type_ = STRING
                token.value = eval(match.group(0))
                return match.group(0)

            if self.scanOperand:
                match = reRegExp.match(input__)
                if match:
                    token.type_ = REGEXP
                    token.value = {"regexp": match.group(1),
                                   "modifiers": match.group(2)}
                    return match.group(0)

            match = opRegExp.match(input__)
            if match:
                op = match.group(0)
                if assignOps.has_key(op) and input__[len(op)] == '=':
                    token.type_ = ASSIGN
                    token.assignOp = globals()[opTypeNames[op]]
                    token.value = op
                    return match.group(0) + "="
                token.type_ = globals()[opTypeNames[op]]
                if self.scanOperand and (token.type_ in (PLUS, MINUS)):
                    token.type_ += UNARY_PLUS - PLUS
                token.assignOp = None
                token.value = op
                return match.group(0)

            if self.scanNewlines:
                match = re.match(r'^\n', input__)
                if match:
                    token.type_ = NEWLINE
                    return match.group(0)

            raise self.newSyntaxError("Illegal token")

        token.start = self.cursor
        self.cursor += len(matchInput())
        token.end = self.cursor
        token.lineno = self.lineno
        return getattr(token, "type_", None)

    def unget(self):
        self.lookahead += 1
        if self.lookahead == 4: raise "PANIC: too much lookahead!"
        self.tokenIndex = (self.tokenIndex - 1) & 3

    def newSyntaxError(self, m):
        return SyntaxError_(m, self.filename, self.lineno)

class CompilerContext(object):
    def __init__(self, inFunction):
        self.inFunction = inFunction
        self.stmtStack = []
        self.funDecls = []
        self.varDecls = []
        self.bracketLevel = 0
        self.curlyLevel = 0
        self.parenLevel = 0
        self.hookLevel = 0
        self.ecmaStrictMode = False
        self.inForLoopInit = False

def Script(t, x):
    n = Statements(t, x)
    n.type_ = SCRIPT
    n.funDecls = x.funDecls
    n.varDecls = x.varDecls
    return n

class Node(list):

    def __init__(self, t, type_=None, args=[]):
        list.__init__(self)

        token = t.token
        if token:
            if type_:
                self.type_ = type_
            else:
                self.type_ = getattr(token, "type_", None)
            self.value = token.value
            self.lineno = token.lineno
            self.start = token.start
            self.end = token.end
        else:
            self.type_ = type_
            self.lineno = t.lineno
        self.tokenizer = t

        for arg in args:
            self.append(arg)

    type = property(lambda self: tokenstr(self.type_))

    # Always use push to add operands to an expression, to update start and end.
    def append(self, kid, numbers=[]):
        if kid:
            if hasattr(self, "start") and kid.start < self.start:
                self.start = kid.start
            if hasattr(self, "end") and self.end < kid.end:
                self.end = kid.end
        return list.append(self, kid)

    indentLevel = 0

    def __str__(self):
        a = list((str(i), v) for i, v in enumerate(self))
        for attr in dir(self):
            if attr[0] == "_": continue
            elif attr == "tokenizer":
                a.append((attr, "[object Object]"))
            elif attr in ("append", "count", "extend", "getSource", "index",
                    "insert", "pop", "remove", "reverse", "sort", "type_",
                    "target", "filename", "indentLevel", "type"):
                continue
            else:
                a.append((attr, getattr(self, attr)))
        if len(self): a.append(("length", len(self)))
        a.sort(lambda a, b: cmp(a[0], b[0]))
        INDENTATION = "    "
        Node.indentLevel += 1
        n = Node.indentLevel
        s = "{\n%stype: %s" % ((INDENTATION * n), tokenstr(self.type_))
        for i, value in a:
            s += ",\n%s%s: " % ((INDENTATION * n), i)
            if i == "value" and self.type_ == REGEXP:
                s += "/%s/%s" % (value["regexp"], value["modifiers"])
            elif value is None:
                s += "null"
            elif value is False:
                s += "false"
            elif value is True:
                s += "true"
            elif type(value) == list:
                s += ','.join((str(x) for x in value))
            else:
                s += str(value)
        Node.indentLevel -= 1
        n = Node.indentLevel
        s += "\n%s}" % (INDENTATION * n)
        return s
    __repr__ = __str__

    def getSource(self):
        if getattr(self, "start", None) is not None:
            if getattr(self, "end", None) is not None:
                return self.tokenizer.source[self.start:self.end]
            return self.tokenizer.source[self.start:]
        if getattr(self, "end", None) is not None:
            return self.tokenizer.source[:self.end]
        return self.tokenizer.source[:]

    filename = property(lambda self: self.tokenizer.filename)

    def __nonzero__(self): return True

# Statement stack and nested statement handler.
def nest(t, x, node, func, end=None):
    x.stmtStack.append(node)
    n = func(t, x)
    x.stmtStack.pop()
    if end: t.mustMatch(end)
    return n

def tokenstr(tt):
    t = tokens[tt]
    if re.match(r'^\W', t):
        return opTypeNames[t]
    return t.upper()

def Statements(t, x):
    n = Node(t, BLOCK)
    x.stmtStack.append(n)
    while not t.done and t.peek() != RIGHT_CURLY:
        n.append(Statement(t, x))
    x.stmtStack.pop()
    return n

def Block(t, x):
    t.mustMatch(LEFT_CURLY)
    n = Statements(t, x)
    t.mustMatch(RIGHT_CURLY)
    return n

DECLARED_FORM = 0
EXPRESSED_FORM = 1
STATEMENT_FORM = 2

def Statement(t, x):
    tt = t.get()

    # Cases for statements ending in a right curly return early, avoiding the
    # common semicolon insertion magic after this switch.
    if tt == FUNCTION:
        if len(x.stmtStack) > 1:
            type_ = STATEMENT_FORM
        else:
            type_ = DECLARED_FORM
        return FunctionDefinition(t, x, True, type_)

    elif tt == LEFT_CURLY:
        n = Statements(t, x)
        t.mustMatch(RIGHT_CURLY)
        return n

    elif tt == IF:
        n = Node(t)
        n.condition = ParenExpression(t, x)
        x.stmtStack.append(n)
        n.thenPart = Statement(t, x)
        if t.match(ELSE):
            n.elsePart = Statement(t, x)
        else:
            n.elsePart = None
        x.stmtStack.pop()
        return n

    elif tt == SWITCH:
        n = Node(t)
        t.mustMatch(LEFT_PAREN)
        n.discriminant = Expression(t, x)
        t.mustMatch(RIGHT_PAREN)
        n.cases = []
        n.defaultIndex = -1
        x.stmtStack.append(n)
        t.mustMatch(LEFT_CURLY)
        while True:
            tt = t.get()
            if tt == RIGHT_CURLY: break

            if tt in (DEFAULT, CASE):
                if tt == DEFAULT and n.defaultIndex >= 0:
                    raise t.newSyntaxError("More than one switch default")
                n2 = Node(t)
                if tt == DEFAULT:
                    n.defaultIndex = len(n.cases)
                else:
                    n2.caseLabel = Expression(t, x, COLON)
            else:
                raise t.newSyntaxError("Invalid switch case")
            t.mustMatch(COLON)
            n2.statements = Node(t, BLOCK)
            while True:
                tt = t.peek()
                if(tt == CASE or tt == DEFAULT or tt == RIGHT_CURLY): break
                n2.statements.append(Statement(t, x))
            n.cases.append(n2)
        x.stmtStack.pop()
        return n

    elif tt == FOR:
        n = Node(t)
        n2 = None
        n.isLoop = True
        t.mustMatch(LEFT_PAREN)
        tt = t.peek()
        if tt != SEMICOLON:
            x.inForLoopInit = True
            if tt == VAR or tt == CONST:
                t.get()
                n2 = Variables(t, x)
            else:
                n2 = Expression(t, x)
            x.inForLoopInit = False

        if n2 and t.match(IN):
            n.type_ = FOR_IN
            if n2.type_ == VAR:
                if len(n2) != 1:
                    raise SyntaxError("Invalid for..in left-hand side",
                            t.filename, n2.lineno)

                # NB: n2[0].type_ == INDENTIFIER and n2[0].value == n2[0].name
                n.iterator = n2[0]
                n.varDecl = n2
            else:
                n.iterator = n2
                n.varDecl = None
            n.object = Expression(t, x)
        else:
            if n2:
                n.setup = n2
            else:
                n.setup = None
            t.mustMatch(SEMICOLON)
            if t.peek() == SEMICOLON:
                n.condition = None
            else:
                n.condition = Expression(t, x)
            t.mustMatch(SEMICOLON)
            if t.peek() == RIGHT_PAREN:
                n.update = None
            else:
                n.update = Expression(t, x)
        t.mustMatch(RIGHT_PAREN)
        n.body = nest(t, x, n, Statement)
        return n

    elif tt == WHILE:
        n = Node(t)
        n.isLoop = True
        n.condition = ParenExpression(t, x)
        n.body = nest(t, x, n, Statement)
        return n

    elif tt == DO:
        n = Node(t)
        n.isLoop = True
        n.body = nest(t, x, n, Statement, WHILE)
        n.condition = ParenExpression(t, x)
        if not x.ecmaStrictMode:
            # <script language="JavaScript"> (without version hints) may need
            # automatic semicolon insertion without a newline after do-while.
            # See http://bugzilla.mozilla.org/show_bug.cgi?id=238945.
            t.match(SEMICOLON)
            return n

    elif tt in (BREAK, CONTINUE):
        n = Node(t)
        if t.peekOnSameLine() == IDENTIFIER:
            t.get()
            n.label = t.token.value
        ss = x.stmtStack
        i = len(ss)
        label = getattr(n, "label", None)
        if label:
            while True:
                i -= 1
                if i < 0:
                    raise t.newSyntaxError("Label not found")
                if getattr(ss[i], "label", None) == label: break
        else:
            while True:
                i -= 1
                if i < 0:
                    if tt == BREAK:
                        raise t.newSyntaxError("Invalid break")
                    else:
                        raise t.newSyntaxError("Invalid continue")
                if (getattr(ss[i], "isLoop", None) or (tt == BREAK and
                        ss[i].type_ == SWITCH)):
                    break
        n.target = ss[i]

    elif tt == TRY:
        n = Node(t)
        n.tryBlock = Block(t, x)
        n.catchClauses = []
        while t.match(CATCH):
            n2 = Node(t)
            t.mustMatch(LEFT_PAREN)
            n2.varName = t.mustMatch(IDENTIFIER).value
            if t.match(IF):
                if x.ecmaStrictMode:
                    raise t.newSyntaxError("Illegal catch guard")
                if n.catchClauses and not n.catchClauses[-1].guard:
                    raise t.newSyntaxError("Gaurded catch after unguarded")
                n2.guard = Expression(t, x)
            else:
                n2.guard = None
            t.mustMatch(RIGHT_PAREN)
            n2.block = Block(t, x)
            n.catchClauses.append(n2)
        if t.match(FINALLY):
            n.finallyBlock = Block(t, x)
        if not n.catchClauses and not getattr(n, "finallyBlock", None):
            raise t.newSyntaxError("Invalid try statement")
        return n

    elif tt in (CATCH, FINALLY):
        raise t.newSyntaxError(tokens[tt] + " without preceding try")

    elif tt == THROW:
        n = Node(t)
        n.exception = Expression(t, x)

    elif tt == RETURN:
        if not x.inFunction:
            raise t.newSyntaxError("Invalid return")
        n = Node(t)
        tt = t.peekOnSameLine()
        if tt not in (END, NEWLINE, SEMICOLON, RIGHT_CURLY):
            n.value = Expression(t, x)

    elif tt == WITH:
        n = Node(t)
        n.object = ParenExpression(t, x)
        n.body = nest(t, x, n, Statement)
        return n

    elif tt in (VAR, CONST):
        n = Variables(t, x)

    elif tt == DEBUGGER:
        n = Node(t)

    elif tt in (NEWLINE, SEMICOLON):
        n = Node(t, SEMICOLON)
        n.expression = None
        return n

    else:
        if tt == IDENTIFIER:
            t.scanOperand = False
            tt = t.peek()
            t.scanOperand = True
            if tt == COLON:
                label = t.token.value
                ss = x.stmtStack
                i = len(ss) - 1
                while i >= 0:
                    if getattr(ss[i], "label", None) == label:
                        raise t.newSyntaxError("Duplicate label")
                    i -= 1
                t.get()
                n = Node(t, LABEL)
                n.label = label
                n.statement = nest(t, x, n, Statement)
                return n

        n = Node(t, SEMICOLON)
        t.unget()
        n.expression = Expression(t, x)
        n.end = n.expression.end

    if t.lineno == t.token.lineno:
        tt = t.peekOnSameLine()
        if tt not in (END, NEWLINE, SEMICOLON, RIGHT_CURLY):
            raise t.newSyntaxError("Missing ; before statement")
    t.match(SEMICOLON)
    return n

def FunctionDefinition(t, x, requireName, functionForm):
    f = Node(t)
    if f.type_ != FUNCTION:
        if f.value == "get":
            f.type_ = GETTER
        else:
            f.type_ = SETTER
    if t.match(IDENTIFIER):
        f.name = t.token.value
    elif requireName:
        raise t.newSyntaxError("Missing function identifier")

    t.mustMatch(LEFT_PAREN)
    f.params = []
    while True:
        tt = t.get()
        if tt == RIGHT_PAREN: break
        if tt != IDENTIFIER:
            raise t.newSyntaxError("Missing formal parameter")
        f.params.append(t.token.value)
        if t.peek() != RIGHT_PAREN:
            t.mustMatch(COMMA)

    t.mustMatch(LEFT_CURLY)
    x2 = CompilerContext(True)
    f.body = Script(t, x2)
    t.mustMatch(RIGHT_CURLY)
    f.end = t.token.end

    f.functionForm = functionForm
    if functionForm == DECLARED_FORM:
        x.funDecls.append(f)
    return f

def Variables(t, x):
    n = Node(t)
    while True:
        t.mustMatch(IDENTIFIER)
        n2 = Node(t)
        n2.name = n2.value
        if t.match(ASSIGN):
            if t.token.assignOp:
                raise t.newSyntaxError("Invalid variable initialization")
            n2.initializer = Expression(t, x, COMMA)
        n2.readOnly = not not (n.type_ == CONST)
        n.append(n2)
        x.varDecls.append(n2)
        if not t.match(COMMA): break
    return n

def ParenExpression(t, x):
    t.mustMatch(LEFT_PAREN)
    n = Expression(t, x)
    t.mustMatch(RIGHT_PAREN)
    return n

opPrecedence = {
    "SEMICOLON": 0,
    "COMMA": 1,
    "ASSIGN": 2, "HOOK": 2, "COLON": 2,
    # The above all have to have the same precedence, see bug 330975.
    "OR": 4,
    "AND": 5,
    "BITWISE_OR": 6,
    "BITWISE_XOR": 7,
    "BITWISE_AND": 8,
    "EQ": 9, "NE": 9, "STRICT_EQ": 9, "STRICT_NE": 9,
    "LT": 10, "LE": 10, "GE": 10, "GT": 10, "IN": 10, "INSTANCEOF": 10,
    "LSH": 11, "RSH": 11, "URSH": 11,
    "PLUS": 12, "MINUS": 12,
    "MUL": 13, "DIV": 13, "MOD": 13,
    "DELETE": 14, "VOID": 14, "TYPEOF": 14,
    # "PRE_INCREMENT": 14, "PRE_DECREMENT": 14,
    "NOT": 14, "BITWISE_NOT": 14, "UNARY_PLUS": 14, "UNARY_MINUS": 14,
    "INCREMENT": 15, "DECREMENT": 15,     # postfix
    "NEW": 16,
    "DOT": 17
}

# Map operator type code to precedence
for i in opPrecedence.copy():
    opPrecedence[globals()[i]] = opPrecedence[i]

opArity = {
    "COMMA": -2,
    "ASSIGN": 2,
    "HOOK": 3,
    "OR": 2,
    "AND": 2,
    "BITWISE_OR": 2,
    "BITWISE_XOR": 2,
    "BITWISE_AND": 2,
    "EQ": 2, "NE": 2, "STRICT_EQ": 2, "STRICT_NE": 2,
    "LT": 2, "LE": 2, "GE": 2, "GT": 2, "IN": 2, "INSTANCEOF": 2,
    "LSH": 2, "RSH": 2, "URSH": 2,
    "PLUS": 2, "MINUS": 2,
    "MUL": 2, "DIV": 2, "MOD": 2,
    "DELETE": 1, "VOID": 1, "TYPEOF": 1,
    # "PRE_INCREMENT": 1, "PRE_DECREMENT": 1,
    "NOT": 1, "BITWISE_NOT": 1, "UNARY_PLUS": 1, "UNARY_MINUS": 1,
    "INCREMENT": 1, "DECREMENT": 1,     # postfix
    "NEW": 1, "NEW_WITH_ARGS": 2, "DOT": 2, "INDEX": 2, "CALL": 2,
    "ARRAY_INIT": 1, "OBJECT_INIT": 1, "GROUP": 1
}

# Map operator type code to arity.
for i in opArity.copy():
    opArity[globals()[i]] = opArity[i]

def Expression(t, x, stop=None):
    operators = []
    operands = []
    bl = x.bracketLevel
    cl = x.curlyLevel
    pl = x.parenLevel
    hl = x.hookLevel

    def reduce_():
        n = operators.pop()
        op = n.type_
        arity = opArity[op]
        if arity == -2:
            # Flatten left-associative trees.
            left = (len(operands) >= 2 and operands[-2])
            if left.type_ == op:
                right = operands.pop()
                left.append(right)
                return left
            arity = 2

        # Always use append to add operands to n, to update start and end.
        a = operands[-arity:]
        del operands[-arity:]
        for operand in a:
            n.append(operand)

        # Include closing bracket or postfix operator in [start,end).
        if n.end < t.token.end:
            n.end = t.token.end

        operands.append(n)
        return n

    class BreakOutOfLoops(Exception): pass
    try:
        while True:
            tt = t.get()
            if tt == END: break
            if (tt == stop and x.bracketLevel == bl and x.curlyLevel == cl and
                    x.parenLevel == pl and x.hookLevel == hl):
                # Stop only if tt matches the optional stop parameter, and that
                # token is not quoted by some kind of bracket.
                break
            if tt == SEMICOLON:
                # NB: cannot be empty, Statement handled that.
                raise BreakOutOfLoops

            elif tt in (ASSIGN, HOOK, COLON):
                if t.scanOperand:
                    raise BreakOutOfLoops
                while ((operators and opPrecedence.get(operators[-1].type_,
                        None) > opPrecedence.get(tt)) or (tt == COLON and
                        operators and operators[-1].type_ == ASSIGN)):
                    reduce_()
                if tt == COLON:
                    if operators:
                        n = operators[-1]
                    if not operators or n.type_ != HOOK:
                        raise t.newSyntaxError("Invalid label")
                    x.hookLevel -= 1
                else:
                    operators.append(Node(t))
                    if tt == ASSIGN:
                        operands[-1].assignOp = t.token.assignOp
                    else:
                        x.hookLevel += 1

                t.scanOperand = True

            elif tt in (IN, COMMA, OR, AND, BITWISE_OR, BITWISE_XOR,
                    BITWISE_AND, EQ, NE, STRICT_EQ, STRICT_NE, LT, LE, GE, GT,
                    INSTANCEOF, LSH, RSH, URSH, PLUS, MINUS, MUL, DIV, MOD,
                    DOT):
                # We're treating comma as left-associative so reduce can fold
                # left-heavy COMMA trees into a single array.
                if tt == IN:
                    # An in operator should not be parsed if we're parsing the
                    # head of a for (...) loop, unless it is in the then part of
                    # a conditional expression, or parenthesized somehow.
                    if (x.inForLoopInit and not x.hookLevel and not
                            x.bracketLevel and not x.curlyLevel and
                            not x.parenLevel):
                        raise BreakOutOfLoops
                if t.scanOperand:
                    raise BreakOutOfLoops
                while (operators and opPrecedence.get(operators[-1].type_)
                        >= opPrecedence.get(tt)):
                    reduce_()
                if tt == DOT:
                    t.mustMatch(IDENTIFIER)
                    operands.append(Node(t, DOT, [operands.pop(), Node(t)]))
                else:
                    operators.append(Node(t))
                    t.scanOperand = True

            elif tt in (DELETE, VOID, TYPEOF, NOT, BITWISE_NOT, UNARY_PLUS,
                    UNARY_MINUS, NEW):
                if not t.scanOperand:
                    raise BreakOutOfLoops
                operators.append(Node(t))

            elif tt in (INCREMENT, DECREMENT):
                if t.scanOperand:
                    operators.append(Node(t)) # prefix increment or decrement
                else:
                    # Don't cross a line boundary for postfix {in,de}crement.
                    if (t.tokens.get((t.tokenIndex + t.lookahead - 1)
                            & 3).lineno != t.lineno):
                        raise BreakOutOfLoops

                    # Use >, not >=, so postfix has higher precedence than
                    # prefix.
                    while (operators and opPrecedence.get(operators[-1].type_,
                            None) > opPrecedence.get(tt)):
                        reduce_()
                    n = Node(t, tt, [operands.pop()])
                    n.postfix = True
                    operands.append(n)

            elif tt == FUNCTION:
                if not t.scanOperand:
                    raise BreakOutOfLoops
                operands.append(FunctionDefinition(t, x, False, EXPRESSED_FORM))
                t.scanOperand = False

            elif tt in (NULL, THIS, TRUE, FALSE, IDENTIFIER, NUMBER, STRING,
                    REGEXP):
                if not t.scanOperand:
                    raise BreakOutOfLoops
                operands.append(Node(t))
                t.scanOperand = False

            elif tt == LEFT_BRACKET:
                if t.scanOperand:
                    # Array initializer. Parse using recursive descent, as the
                    # sub-grammer here is not an operator grammar.
                    n = Node(t, ARRAY_INIT)
                    while True:
                        tt = t.peek()
                        if tt == RIGHT_BRACKET: break
                        if tt == COMMA:
                            t.get()
                            n.append(None)
                            continue
                        n.append(Expression(t, x, COMMA))
                        if not t.match(COMMA):
                            break
                    t.mustMatch(RIGHT_BRACKET)
                    operands.append(n)
                    t.scanOperand = False
                else:
                    operators.append(Node(t, INDEX))
                    t.scanOperand = True
                    x.bracketLevel += 1

            elif tt == RIGHT_BRACKET:
                if t.scanOperand or x.bracketLevel == bl:
                    raise BreakOutOfLoops
                while reduce_().type_ != INDEX:
                    continue
                x.bracketLevel -= 1

            elif tt == LEFT_CURLY:
                if not t.scanOperand:
                    raise BreakOutOfLoops
                # Object initializer. As for array initializers (see above),
                # parse using recursive descent.
                x.curlyLevel += 1
                n = Node(t, OBJECT_INIT)

                class BreakOutOfObjectInit(Exception): pass
                try:
                    if not t.match(RIGHT_CURLY):
                        while True:
                            tt = t.get()
                            if ((t.token.value == "get" or
                                    t.token.value == "set") and
                                    t.peek == IDENTIFIER):
                                if x.ecmaStrictMode:
                                    raise t.newSyntaxError("Illegal property "
                                            "accessor")
                                n.append(FunctionDefinition(t, x, True,
                                        EXPRESSED_FORM))
                            else:
                                if tt in (IDENTIFIER, NUMBER, STRING):
                                    id_ = Node(t)
                                elif tt == RIGHT_CURLY:
                                    if x.ecmaStrictMode:
                                        raise t.newSyntaxError("Illegal "
                                                "trailing ,")
                                    raise BreakOutOfObjectInit
                                else:
                                    raise t.newSyntaxError("Invalid property "
                                            "name")
                                t.mustMatch(COLON)
                                n.append(Node(t, PROPERTY_INIT, [id_,
                                        Expression(t, x, COMMA)]))
                            if not t.match(COMMA): break
                        t.mustMatch(RIGHT_CURLY)
                except BreakOutOfObjectInit, e: pass
                operands.append(n)
                t.scanOperand = False
                x.curlyLevel -= 1

            elif tt == RIGHT_CURLY:
                if not t.scanOperand and x.curlyLevel != cl:
                    raise ParseError("PANIC: right curly botch")
                raise BreakOutOfLoops

            elif tt == LEFT_PAREN:
                if t.scanOperand:
                    operators.append(Node(t, GROUP))
                    x.parenLevel += 1
                else:
                    while (operators and
                            opPrecedence.get(operators[-1].type_) >
                            opPrecedence[NEW]):
                        reduce_()

                    # Handle () now, to regularize the n-ary case for n > 0.
                    # We must set scanOperand in case there are arguments and
                    # the first one is a regexp or unary+/-.
                    if operators:
                        n = operators[-1]
                    else:
                        n = Object()
                        n.type_ = None
                    t.scanOperand = True
                    if t.match(RIGHT_PAREN):
                        if n.type_ == NEW:
                            operators.pop()
                            n.append(operands.pop())
                        else:
                            n = Node(t, CALL, [operands.pop(), Node(t, LIST)])
                        operands.append(n)
                        t.scanOperand = False
                    else:
                        if n.type_ == NEW:
                            n.type_ = NEW_WITH_ARGS
                        else:
                            operators.append(Node(t, CALL))
                        x.parenLevel += 1

            elif tt == RIGHT_PAREN:
                if t.scanOperand or x.parenLevel == pl:
                    raise BreakOutOfLoops
                while True:
                    tt = reduce_().type_
                    if tt in (GROUP, CALL, NEW_WITH_ARGS):
                        break
                if tt != GROUP:
                    if operands:
                        n = operands[-1]
                        if n[1].type_ != COMMA:
                            n[1] = Node(t, LIST, [n[1]])
                        else:
                            n[1].type_ = LIST
                    else:
                        raise ParseError, "Unexpected amount of operands"
                x.parenLevel -= 1

            # Automatic semicolon insertion means we may scan across a newline
            # and into the beginning of another statement. If so, break out of
            # the while loop and let the t.scanOperand logic handle errors.
            else:
                raise BreakOutOfLoops
    except BreakOutOfLoops, e: pass

    if x.hookLevel != hl:
        raise t.newSyntaxError("Missing : after ?")
    if x.parenLevel != pl:
        raise t.newSyntaxError("Missing ) in parenthetical")
    if x.bracketLevel != bl:
        raise t.newSyntaxError("Missing ] in index expression")
    if t.scanOperand:
        raise t.newSyntaxError("Missing operand")

    t.scanOperand = True
    t.unget()
    while operators:
        reduce_()
    return operands.pop()

def parse(source, filename=None, starting_line_number=1):
    """Parse some Javascript

    Args:
        source: the Javascript source, as a string
        filename: the filename to include in messages
        starting_line_number: the line number of the first line of the
            passed in source, for output messages
    Returns:
        the parsed source code data structure
    Raises:
        ParseError
    """
    t = Tokenizer(source, filename, starting_line_number)
    x = CompilerContext(False)
    n = Script(t, x)
    if not t.done:
        raise t.newSyntaxError("Syntax error")
    return n

if __name__ == "__main__":
    print str(parse(file(sys.argv[1]).read(),sys.argv[1]))



# -*- coding: utf-8 -*-

ENG, RUS = range(1, 3)
LANGUAGE = ('English', 'Russian')

def get_lang_dict(language):
    if language == ENG:
        import english as lang
    elif language == RUS:
        import russian as lang
    return lang.dictionary



# -*- coding: utf-8 -*-

import sys
import os
import datetime, time
import xml.dom.minidom
import cPickle
import locale
import decimal
import zipfile
import gzip
import cStringIO

DEFAULT_SETTINGS = {
    'LANGUAGE': 1,
    'SAFE_MODE': False,
    'DEBUGGING': False,
    'LOG_FILE': '',
    'DECIMAL_POINT': '.',
    'MON_DECIMAL_POINT': '.',
    'MON_THOUSANDS_SEP': '',
    'CURRENCY_SYMBOL': '',
    'FRAC_DIGITS': 2,
    'P_CS_PRECEDES': False,
    'N_CS_PRECEDES': False,
    'P_SEP_BY_SPACE': True,
    'N_SEP_BY_SPACE': True,
    'POSITIVE_SIGN': '',
    'NEGATIVE_SIGN': '-',
    'P_SIGN_POSN': 1,
    'N_SIGN_POSN': 1,
    'D_FMT': '%x',
    'D_T_FMT': '%X',
    'CON_POOL_SIZE': 4
}

LOCALE_SETTINGS = (
    'DECIMAL_POINT',
    'MON_DECIMAL_POINT',
    'MON_THOUSANDS_SEP',
    'CURRENCY_SYMBOL',
    'FRAC_DIGITS',
    'P_CS_PRECEDES',
    'N_CS_PRECEDES',
    'P_SEP_BY_SPACE',
    'N_SEP_BY_SPACE',
    'POSITIVE_SIGN',
    'NEGATIVE_SIGN',
    'P_SIGN_POSN',
    'N_SIGN_POSN',
    'D_FMT',
    'D_T_FMT'
)

SETTINGS = {}

VERSION = '1.0'
RESPONSE, NOT_LOGGED, UNDER_MAINTAINANCE = range(1, 4)
SQLITE, FIREBIRD, POSTGRESQL, MYSQL = range(1, 5)
DB_TYPE = ('Sqlite', 'FireBird', 'PostgreSQL')
#DB_TYPE = ('Sqlite', 'FireBird', 'PostgreSQL', 'mySQL')
ROOT_TYPE, USERS_TYPE, ROLES_TYPE, TASKS_TYPE, TASK_TYPE, \
    CATALOGS_TYPE, JOURNALS_TYPE, TABLES_TYPE, REPORTS_TYPE, \
    CATALOG_TYPE, JOURNAL_TYPE, TABLE_TYPE, REPORT_TYPE, DETAIL_TYPE = range(1, 15)

ITEM_TYPES = ["root", "users", "roles", "tasks", 'task',
        "catalogs", "journals", "tables", "reports",
        "catalog", "journal", "table", "report", "detail"]

SYSTEM_FIELDS = KEY_FIELD_NAME, DEL_FIELD_NAME, OWNER_ID_FIELD_NAME, OWNER_REC_ID_FIELD_NAME = 'id', 'deleted', 'owner_id', 'owner_rec_id'
TEXT, INTEGER, FLOAT, CURRENCY, DATE, DATETIME, BOOLEAN, BLOB = range(1, 9)
FIELD_TYPES = ('TEXT', 'INTEGER', 'FLOAT', 'CURRENCY', 'DATE', 'DATETIME', 'BOOLEAN', 'BLOB')
FIELD_TYPE_NAMES = ('', 'text', 'integer', 'float', 'currency', 'date', 'datetime', 'boolean', 'blob')
ALIGN_LEFT, ALIGN_CENTER, ALIGN_RIGHT = 1, 2, 3
ALIGNMENT = ('ALIGN_LEFT', 'ALIGN_CENTER', 'ALIGN_RIGHT')
SYS, DATA, order = range(3)

FILTER_EQ, FILTER_NE, FILTER_LT, FILTER_LE, FILTER_GT, FILTER_GE, FILTER_IN, FILTER_NOT_IN, \
FILTER_RANGE, FILTER_ISNULL, FILTER_EXACT, FILTER_CONTAINS, FILTER_STARTWITH, FILTER_ENDWITH, \
FILTER_SEARCH = range(1, 16)
FILTER_STR = ('eq', 'ne', 'lt', 'le', 'gt', 'ge', 'in', 'not_in', \
'range', 'isnull', 'exact', 'contains', 'startwith', 'endwith', \
'search')
FILTER_SIGN = ('', '=', '<>', '<', '<=', '>', '>=', 'IN', 'NOT IN',
    'BETWEEN', 'ISNULL', '=', 'LIKE', 'LIKE', 'LIKE', 'SEARCH')
FILTER_STRING = ('EQ', 'NE', 'LT', 'LE', 'GT', 'GE', 'IN', 'NOT IN',
    'RANGE', 'ISNULL', 'EXACT', 'CONTAINS', 'STARTWITH', 'ENDWITH', 'SEARCH')
REC_STATUS, REC_CONTROLS_INFO, REC_CHANGE_ID = range(3)

ORDER_ASC, ORDER_DESC = range(2)
STATE_NONE, STATE_BROWSE, STATE_INSERT, STATE_EDIT, STATE_DELETE = range(5)
UPDATE_OPEN, UPDATE_DELETE, UPDATE_CANCEL, UPDATE_APPEND, UPDATE_INSERT, UPDATE_SCROLLED, UPDATE_RESTORE, UPDATE_REFRESH = range(8)
RECORD_UNCHANGED, RECORD_INSERTED, RECORD_MODIFIED, RECORD_DETAILS_MODIFIED, RECORD_DELETED = None, 1, 2, 3, 4
EDITOR_TAB_SIZE = 4
ITEM_PARAM_INDENT = '__$_item_$__'
FIELD_PARAM_INDENT = '__$_field_$__'
FILTER_PARAM_INDENT = '__$_filter_$__'

CLIENT_MODULE, WEB_CLIENT_MODULE, SERVER_MODULE = range(3)
TAB_FUNCS, TAB_EVENTS, TAB_TASK, TAB_FIELDS = range(4)
editor_tabs = ("Module", "Events", "Task", "Fields")

mime_types = {
    'text/html':                             ['html', 'htm', 'shtml'],
    'text/css':                              ['css'],
    'text/xml':                              ['xml'],
    'image/gif':                             ['gif'],
    'image/jpeg':                            ['jpeg', 'jpg'],
    'application/x-javascript':              ['js'],
    'application/atom+xml':                  ['atom'],
    'application/rss+xml':                   ['rss'],

    'text/mathml':                           ['mml'],
    'text/plain':                            ['txt'],
    'text/vnd.sun.j2me.app-descriptor':      ['jad'],
    'text/vnd.wap.wml':                      ['wml'],
    'text/x-component':                      ['htc'],

    'image/png':                             ['png'],
    'image/tiff':                            ['tif', 'tiff'],
    'image/vnd.wap.wbmp':                    ['wbmp'],
    'image/x-icon':                          ['ico'],
    'image/x-jng':                           ['jng'],
    'image/x-ms-bmp':                        ['bmp'],
    'image/svg+xml':                         ['svg'],
    'image/webp':                            ['webp'],

    'application/java-archive':              ['jar', 'war', 'ear'],
    'application/mac-binhex40':              ['hqx'],
    'application/msword':                    ['doc'],
    'application/pdf':                       ['pdf'],
    'application/postscript':                ['ps', 'eps', 'ai'],
    'application/rtf':                       ['rtf'],
    'application/vnd.ms-excel':              ['xls'],
    'application/vnd.ms-powerpoint':         ['ppt'],
    'application/vnd.wap.wmlc':              ['wmlc'],
    'application/vnd.google-earth.kml+xml':  ['kml'],
    'application/vnd.google-earth.kmz':      ['kmz'],
    'application/x-7z-compressed':           ['7z'],
    'application/x-cocoa':                   ['cco'],
    'application/x-java-archive-diff':       ['jardiff'],
    'application/x-java-jnlp-file':          ['jnlp'],
    'application/x-makeself':                ['run'],
    'application/x-perl':                    ['pl', 'pm'],
    'application/x-pilot':                   ['prc', 'pdb'],
    'application/x-rar-compressed':          ['rar'],
    'application/x-redhat-package-manager':  ['rpm'],
    'application/x-sea':                     ['sea'],
    'application/x-shockwave-flash':         ['swf'],
    'application/x-stuffit':                 ['sit'],
    'application/x-tcl':                     ['tcl', 'tk'],
    'application/x-x509-ca-cert':            ['der', 'pem', 'crt'],
    'application/x-xpinstall':               ['xpi'],
    'application/xhtml+xml':                 ['xhtml'],
    'application/zip':                       ['zip'],

    'application/octet-stream':              ['bin', 'exe', 'dll'],
    'application/octet-stream':              ['deb'],
    'application/octet-stream':              ['dmg'],
    'application/octet-stream':              ['eot'],
    'application/octet-stream':              ['iso', 'img'],
    'application/octet-stream':              ['msi', 'msp', 'msm'],

    'audio/midi':                            ['mid', 'midi', 'kar'],
    'audio/mpeg':                            ['mp3'],
    'audio/ogg':                             ['ogg'],
    'audio/x-realaudio':                     ['ra'],

    'video/3gpp':                            ['3gpp', '3gp'],
    'video/mpeg':                            ['mpeg', 'mpg'],
    'video/quicktime':                       ['mov'],
    'video/x-flv':                           ['flv'],
    'video/x-mng':                           ['mng'],
    'video/x-ms-asf':                        ['asx', 'asf'],
    'video/x-ms-wmv':                        ['wmv'],
    'video/x-msvideo':                       ['avi'],
    'video/mp4':                             ['m4v', 'mp4']
}

ext_mime_types = {}
for mime, ext in mime_types.items():
    for e in ext:
        ext_mime_types[e] = mime

def mime_type_by_ext(ext):
    ext = ext.replace('.', '')
    return ext_mime_types.get(ext)

def get_alignment(data_type, item=None, value_list=None):
    if (data_type == INTEGER) or (data_type == FLOAT) or (data_type == CURRENCY):
        result = ALIGN_RIGHT
    elif (data_type == DATE) or (data_type == DATETIME):
        result = ALIGN_CENTER
    else:
        result = ALIGN_LEFT
    if item or value_list:
        result = ALIGN_LEFT
    return result

def float_to_str(val):
    return str(val).replace('.', DECIMAL_POINT)

def str_to_float(val):
    val = val.replace(DECIMAL_POINT, '.')
    return float(val)

def cur_to_str(value):

    def transform_digits(val):
        if not val[0].isdigit():
            val = val[1:]
        point = val.find('.')
        dec = ''
        digits = val
        if point >= 0:
            dec = val[point + 1:]
            digits = val[:point]
        result = ''
        count = 0
        lenth = len(digits)
        for i in range(lenth):
            d = digits[lenth - i - 1]
            result = d + result
            count += 1
            if count % 3 == 0 and (i != lenth - 1):
                result = MON_THOUSANDS_SEP + result
        if dec:
            result = result + MON_DECIMAL_POINT + dec
        return result

    if value is None:
        value = 0
    format_str = '%.' + str(FRAC_DIGITS) + 'f'
    result = format_str % value
    result = transform_digits(result)
    if value < 0:
        if N_SIGN_POSN == 3:
            result = NEGATIVE_SIGN + result
        elif N_SIGN_POSN == 4:
            result = result + NEGATIVE_SIGN
    else:
        if P_SIGN_POSN == 3:
            result = POSITIVE_SIGN + result
        elif P_SIGN_POSN == 4:
            result = result + POSITIVE_SIGN
    if CURRENCY_SYMBOL:
        if value < 0:
            if N_CS_PRECEDES:
                if N_SEP_BY_SPACE:
                    result = CURRENCY_SYMBOL + ' ' + result
                else:
                    result = CURRENCY_SYMBOL + result
            else:
                if N_SEP_BY_SPACE:
                    result = result + ' ' + CURRENCY_SYMBOL
                else:
                    result = result + CURRENCY_SYMBOL
        else:
            if P_CS_PRECEDES:
                if P_SEP_BY_SPACE:
                    result = CURRENCY_SYMBOL + ' ' + result
                else:
                    result = CURRENCY_SYMBOL + result
            else:
                if P_SEP_BY_SPACE:
                    result = result + ' ' + CURRENCY_SYMBOL
                else:
                    result = result + CURRENCY_SYMBOL
    if value < 0:
        if N_SIGN_POSN == 0 and NEGATIVE_SIGN:
            result = NEGATIVE_SIGN + '(' + result + ')'
        elif N_SIGN_POSN == 1:
            result = NEGATIVE_SIGN + result
        elif N_SIGN_POSN == 2:
            result = result + NEGATIVE_SIGN
    else:
        if P_SIGN_POSN == 0 and POSITIVE_SIGN:
            result = POSITIVE_SIGN + '(' + result + ')'
        elif P_SIGN_POSN == 1:
            result = POSITIVE_SIGN + result
        elif P_SIGN_POSN == 2:
            result = result + POSITIVE_SIGN
    return result

def currency_to_str(val):
    return cur_to_str(val)

def str_to_currency(val):
    result = val.strip()
    if MON_THOUSANDS_SEP:
        result = result.replace(MON_THOUSANDS_SEP, '')
    if CURRENCY_SYMBOL:
        result = result.replace(CURRENCY_SYMBOL, '')
    if POSITIVE_SIGN:
        result = result.replace(POSITIVE_SIGN, '')
    if NEGATIVE_SIGN and result.find(NEGATIVE_SIGN) != -1:
        result = result.replace(NEGATIVE_SIGN, '')
        result = '-' + result
    result = result.replace(MON_DECIMAL_POINT, '.').strip()
    result = float(result)
    return result

def date_to_str(date):
    return date.strftime(D_FMT)

def str_to_date(date_str):
    time_tuple = time.strptime(date_str, D_FMT)
    return datetime.date(time_tuple.tm_year, time_tuple.tm_mon, time_tuple.tm_mday)

def datetime_to_str(date):
    return date.strftime(D_T_FMT)

def str_to_datetime(date_str):
    time_tuple = time.strptime(date_str, D_T_FMT)
    return datetime.datetime(time_tuple.tm_year, time_tuple.tm_mon,
        time_tuple.tm_mday, time_tuple.tm_hour, time_tuple.tm_min, time_tuple.tm_sec)

def ui_to_string(file_name):
    with open(file_name, "r") as f:
        return f.read()

def load_interface(item):
    item._view_list = []
    item._edit_list = []
    item._order_list = []
    item._reports_list = []
    if item.f_info.value:
        lists = cPickle.loads(str(item.f_info.value))
        item._view_list = lists['view']
        item._edit_list = lists['edit']
        item._order_list = lists['order']
        if lists.get('reports'):
            item._reports_list = lists['reports']

def store_interface(item):
    handlers = item.store_handlers()
    item.clear_handlers()
    try:
        item.edit()
        dic = {'view': item._view_list,
                'edit': item._edit_list,
                'order': item._order_list,
                'reports': item._reports_list}
        item.f_info.value = cPickle.dumps(dic)
        item.post()
        item.apply()
    finally:
        handlers = item.load_handlers(handlers)

def valid_identifier(name):
    if name[0].isdigit():
        return False
    try:
        while vars().get(name):
            name += '1'
        vars()[name] = 1
        eval(name)
        return True
    except:
        return False

def empty_strings(text, module_type):

    def find_start(line, strs, start):
        pos = len(line)
        char = ''
        for c in strs:
            p = line.find(c, start)
            if p != -1 and p < pos:
                pos = p
                char = c
        if pos == len(line):
            pos = -1
        return pos, char


    if module_type == WEB_CLIENT_MODULE:
        strings = ("'", '"')
    else:
        strings = ('"""', "'''", "'", '"')

    result = []
    comment = False
    for line in text.splitlines(True):
        if comment:
            pos = line.find(str_ch)
            if pos != -1:
                comment = False
                line = pos * ' ' + str_ch + line[pos + len(str_ch):]
            else:
                line = ' ' * len(line)
        else:
            start = 0
            while True:
                pos, str_ch = find_start(line, strings, start)
                if pos != -1:
                    end = line.find(str_ch, pos + len(str_ch))
                    if end != -1:
                        line = line[0:pos] + str_ch + ' ' * (end - pos - len(str_ch)) + line[end:]
                        start = end + len(str_ch) + 1
                    else:
                        comment = True
                        line = line[0:pos + len(str_ch)] + ' ' * (len(line) - (pos + len(str_ch)))
                        break;
                else:
                    break;
        result.append(line)
    result = ''.join(result)
    return result

def remove_comments(text, module_type, comment_sign):
    result = []
    if text:
        comment = False
        for line in text.splitlines(True):
            if comment:
                pos = line.find('*/')
                if pos != -1:
                    comment = False
                    line = pos * ' ' + '*/' + line[pos + 2:]
                else:
                    line = ' ' * len(line)
            else:
                pos = line.find(comment_sign)
                if pos != -1:
                    line = line[0:pos] + comment_sign + (len(line) - len(line[0:pos] + comment_sign) - 1) * ' ' + '\n'
                if module_type == WEB_CLIENT_MODULE:
                    pos = line.find('/*')
                    if pos != -1:
                        end = line.find('*/', pos + 2)
                        if end != -1:
                            line = line[0:pos] + '/*' + ' ' * (end - pos - 2) + line[end:]
                        else:
                            comment = True
                            line = line[0:pos+2] + ' ' * (len(line) - (pos + 2))
            result.append(line)
        result = ''.join(result)
    return result

def get_funcs_info(text, module_type):

    def check_line(line, comment_sign, func_literal):
        func_name = ''
        trimed_line = line.strip()
        if len(trimed_line) > 0:
            if not (trimed_line[:len(comment_sign)] == comment_sign):
                indent = line.find(func_literal)
                if indent >= 0:
                    def_end = line.find('(')
                    if def_end > indent:
                        func_name = line[indent+len(func_literal):def_end].strip()
                        if func_name:
                            return (indent, func_name)

    def add_child_funcs(i, parent_indent, parent_dic, parent_key):
        dic = {}
        parent_dic[parent_key] = dic
        if i < len(funcs_list):
            cur_indent = funcs_list[i][0]
        else:
            return
        cur_indent = -1
        child_indent = -1
        while i < len(funcs_list):
            (indent, func_name) = funcs_list[i]
            if cur_indent == -1:
                cur_indent = indent
            if indent == cur_indent:
                dic[func_name] = None
                cur_func_name = func_name
            elif indent > cur_indent:
                if child_indent == -1:
                    child_indent = indent
                if not indent > child_indent:
                    i = add_child_funcs(i, indent, dic, cur_func_name)
            elif indent < cur_indent:
                return i - 1
            i += 1
        return i

    funcs = {}
    funcs['result'] = {}
    if text:
        if module_type == WEB_CLIENT_MODULE:
            comment_sign = '//'
            func_literal = 'function'
        else:
            comment_sign = '#'
            func_literal = 'def'
        text = remove_comments(text, module_type, comment_sign)
        lines = text.splitlines()
        funcs_list = []
        for i, line in enumerate(lines):
            res = check_line(line, comment_sign, func_literal)
            if res:
                funcs_list.append(res)
        add_child_funcs(0, -1, funcs, 'result')
    return funcs['result']

class SingleInstance(object):
    def __init__(self, port=None):
        import sqlite3
        file_name =  os.path.basename(sys.modules['__main__'].__file__)
        self.pid_file = '%s.pid' % file_name
        if port:
            self.pid_file = '%s_%s.pid' % (file_name, port)
        self.con = sqlite3.connect(self.pid_file, timeout=0.1, isolation_level='EXCLUSIVE')
        self.cur = self.con.cursor()
        try:
            self.cur.execute('PRAGMA journal_mode = MEMORY;')
            self.cur.execute('PRAGMA synchronous = OFF;')
            self.cur.execute('CREATE TABLE IF NOT EXISTS PID (ID INTEGER NOT NULL)')
            self.con.commit()
            self.cur.execute('INSERT INTO PID (ID) VALUES (?)', (1,))
        except sqlite3.OperationalError, e:
            if e.args[0].lower().find('database is locked') != -1:
                self.con.close()
                if port:
                    print '%s port %s: another instance is already running, quitting' % (file_name, port)
                else:
                    print '%s: another instance is already running, quitting' % file_name
                sys.exit(-1)

    def close(self):
        if os.path.exists(self.pid_file):
            os.remove(self.pid_file)

def json_defaul_handler(obj):
    result = obj
    if hasattr(obj, 'isoformat'):
        result = obj.isoformat()
    elif isinstance(obj, decimal.Decimal):
        result = float(obj)
    return result

def zip_dir(dir, zip_file, exclude_dirs=[], exclude_ext=[]):
    direct = os.path.join(os.getcwd().decode('utf-8'), dir)
    for dirpath, dirnames, filenames in os.walk(direct):
        head, tail = os.path.split(dirpath)
        if not tail in exclude_dirs:
            for file_name in filenames:
                name, ext = os.path.splitext(file_name)
                if not ext in exclude_ext:
                    file_path = os.path.join(dirpath, file_name)
                    arcname = os.path.relpath(os.path.join(dir, file_path))
                    zip_file.write(file_path, arcname)

def unzip_dir(dir, zip_file_name):
    z = zipfile.ZipFile(zip_file_name)
    for file_name in z.namelist():
        path = os.path.join(dir, os.path.join(*file_name.split('/')))
        dir_name = os.path.dirname(path)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        data = z.read(file_name)
        with open(path, 'w') as f:
            f.write(data)

def now():
    return datetime.datetime.now()

def min_diff(diff):
    return divmod(diff.days * 86400 + diff.seconds, 60)[0]

def hour_diff(diff):
    return divmod(diff.days * 86400 + diff.seconds, 3600)[0]

def compressBuf(buf):
    zbuf = cStringIO.StringIO()
    zfile = gzip.GzipFile(mode = 'wb',  fileobj = zbuf, compresslevel = 9)
    zfile.write(buf)
    zfile.close()
    return zbuf.getvalue()



# -*- coding: utf-8 -*-

from __future__ import with_statement
import sys, os
if not hasattr(sys, 'db_multiprocessing'):
    sys.db_multiprocessing = False

import Queue
import multiprocessing
#from multiprocessing.managers import BaseManager
import zipfile
from xml.dom.minidom import parseString
from xml.sax.saxutils import escape
import datetime, time
import traceback
import inspect

import common
from items import *
from dataset import *
from sql import *

class ServerField(DBField):
    def __init__(self, owner, field_kind, field_name, caption, data_type, required = False,
            item = None, object_field = None, visible = True, index=0,
            edit_visible = True, edit_index = 0, read_only = False,
            expand = False, word_wrap = False, size = 0, default = False, calculated = False,
            editable = False, master_field = None, value_list=None):
        DBField.__init__(self)
        self.owner = owner
        self.field_kind = field_kind
        self.field_name = field_name
        self.field_caption = caption
        self.field_size = size
        self.is_default = default
        self.data_type = data_type
        self.required = required
        self.lookup_field = None
        self.lookup_item = None
        self.master_field = None
        if item:
            self.lookup_item = item
            self.lookup_field = object_field
        if master_field:
            self.master_field = master_field
            self.lookup_field = object_field
        self.view_visible = visible
        self.view_index = index
        self.edit_visible = edit_visible
        self.edit_index = edit_index
        self.read_only = read_only
        self.expand = expand
        self.word_wrap = word_wrap
        self.calculated = calculated
        self.editable = editable
        self.value_list = value_list

    def copy(self, owner):
        result = ServerField(owner, self.field_kind, self.field_name, self.field_caption, self.data_type)
        result.set_info(self.get_info())
        result.lookup_item = self.lookup_item
        return result

class ServerFilterField(DBField):
    def __init__(self, fltr, field):
        DBField.__init__(self)
        self.owner = field.owner
        self.filter = fltr
        self.lookup_item = None
        info = field.get_info()
        self.set_info(field.get_info())

    def do_before_changed(self, new_value, new_lookup_value):
        pass

    def get_row(self):
        return self.owner._filter_row

    def check_reqired(self, value):
        return True

    def set_modified(self, value):
        pass

    def set_record_status(self, value):
        pass

class ServerFilter(DBFilter):
    def __init__(self, owner, name, caption, field_name, filter_type = common.FILTER_EQ, data_type = None, visible = True):
        DBFilter.__init__(self)
        self.owner = owner
        self.filter_name = name
        self.filter_caption = caption
        self.field_name = None
        self.field_ID = None
        if type(field_name) == int:
            self.field_name = owner._field_by_ID(field_name).field_name
        else:
            self.field_name = field_name
        self.filter_type = filter_type
        self.data_type = data_type;
        self.visible = visible
        self.list = []
        if self.field_name:
            field = self.owner._field_by_name(self.field_name)
            self.field = ServerFilterField(self, field)
            setattr(self, self.field_name, self.field)

    def copy(self, owner):
        result = ServerFilter(owner, self.filter_name, self.filter_caption, self.field_name, self.filter_type, self.visible)
        return result


class ServerDataset(Dataset, SQL):
    def __init__(self, table_name='', view_template='', edit_template='', filter_template='', soft_delete=True):
        Dataset.__init__(self)
        self.ID = None
        self.table_name = table_name
        self.view_template = view_template
        self.edit_template = edit_template
        self.filter_template = filter_template
        self.filter_template = filter_template
        self._order_by = []
        self.on_get_next_id = None
        self.values = None
        self.on_select = None
        self.on_apply = None
        self.on_record_count = None
        self.on_get_field_text = None
        self.id_field_name = None
        self.deleted_field_name = None
        self.soft_delete = soft_delete

    def copy(self, filters=True, details=True, handlers=True):
        result = super(ServerDataset, self).copy(filters, details, handlers)
        result.table_name = self.table_name
        result._order_by = self._order_by
        result.id_field_name = self.id_field_name
        result.deleted_field_name = self.deleted_field_name
        result.soft_delete = self.soft_delete
        return result

    def get_event(self, caption):
        return getattr(caption)

    def add_field(self, field_kind, field_name, caption, data_type, required = False,
        item = None, object_field = None,
        visible = True, index=0, edit_visible = True, edit_index = 0, read_only = False, expand = False,
        word_wrap = False, size = 0, default = False, calculated = False, editable = False, master_field = None, value_list=None):
        field = ServerField(self, field_kind, field_name, caption, data_type, required, item, object_field, visible,
            index, edit_visible, edit_index, read_only, expand, word_wrap, size, default, calculated, editable, master_field, value_list)
        self._fields.append(field)
        return field

    def add_filter(self, name, caption, field_name, filter_type = common.FILTER_EQ, data_type = None, visible = True):
        fltr = ServerFilter(self, name, caption, field_name, filter_type, data_type, visible)
        self.filters.append(fltr)
        fltr.owner = self
        setattr(self.filters, name, fltr)
        return fltr

    def __getattr__(self, name):
        if self.detail_by_name(name):
            obj = self.detail_by_name(name)
            setattr(self, name, obj)
            return obj
        else:
            raise AttributeError(self.item_name + ' attribute: ' + name)

    def do_internal_open(self, params):
        return self.select_records(params)

    def do_apply(self, params=None):
        result = True
        if not self.master and self.log_changes:
            if self.item_state != common.STATE_BROWSE:
                raise Exception, u'Item: %s is not in browse state. Apply requires browse state.'
            changes = {}
            self.change_log.get_changes(changes)
            if changes['data']:
                data = self.apply_changes((changes, params),
                    {'can_view': True, 'can_create': True, 'can_edit': True, 'can_delete': True})
                if data:
                    if data['error']:
                        raise Exception, data['error']
                    else:
                        self.change_log.update(data['result'])
        return result

    def get_details(self):
        return self.details

    def get_fields_info(self):
        result = []
        for field in self._fields:
            result.append(field.get_info())
        return result

    def get_filters_info(self):
        result = []
        for fltr in self.filters:
            result.append(fltr.get_info())
        return result

    def get_details_info(self):
        result = []
        for detail in self.details:
            result.append(detail.get_info())
        return result

    #~ def get_reports_info(self):
        #~ result = []
        #~ for report in self.reports:
            #~ result.append(report.ID)
        #~ return result

    def add_detail(self, table):
        detail = ServerDetail(self, table.item_name, table.item_caption, table.table_name)
        self.details.append(detail)
        detail.owner = self
        detail.init_fields()
        return detail

    def detail_by_name(self, caption):
        for table in self.details:
            if table.item_name == caption:
                return table

    def do_on_loaded(self):
        self._filter_row = []
        for i, fltr in enumerate(self.filters):
            self._filter_row.append(None)
            fltr.field.bind_index = i

    def get_view_template(self):
        if self.view_template:
            return self.view_template
        elif self.owner:
            return self.owner.get_view_template()

    def get_view_ui(self):
        return common.ui_to_string(os.path.join(self.task.get_ui_path(), 'ui', self.get_view_template()))

    def get_edit_template(self):
        if self.edit_template:
            return self.edit_template
        elif self.owner:
            return self.owner.get_edit_template()

    def get_edit_ui(self):
        return common.ui_to_string(os.path.join(self.task.get_ui_path(), 'ui', self.get_edit_template()))

    def get_filter_template(self):
        if self.filter_template:
            return self.filter_template
        elif self.owner:
            return self.owner.get_filter_template()

    def get_filter_ui(self):
        if self.owner.filter_template:
            return common.ui_to_string(os.path.join(self.task.get_ui_path(), 'ui', self.get_filter_template()))

    def change_order(self, *fields):
        self._order_by = []
        for field in fields:
            field_name = field
            desc = False
            if field[0] == '-':
                desc = True
                field_name = field[1:]
            try:
                fld = self._field_by_name(field_name)
            except:
                raise RuntimeError('%s: change_order method arument error - %s' % (self.item_name, field))
            self._order_by.append([fld.ID, desc])
        return self

    def get_record_count(self, params, user_info=None, enviroment=None):
        result = 0;
        if self.on_record_count:
            result, error_mes = self.on_record_count(self, params, user_info, enviroment)
        else:
            error_mes = ''
            result = 0
            sql = self.get_record_count_query(params)
            try:
                rows = self.task.execute_select(sql)
                result = rows[0][0]
            except Exception, e:
                error_mes = e.message
        return result, error_mes

    def select_records(self, params, user_info=None, enviroment=None):
        if self.on_select:
            rows, error_mes = self.on_select(self, params, user_info, enviroment)
        else:
            sql = self.get_select_statement(params)
            error_mes = ''
            rows = []
            try:
                rows = self.task.execute_select(sql)
            except Exception, e:
                error_mes = e.message
        return rows, error_mes

    def apply_changes(self, data, privileges, user_info=None, enviroment=None):
        error = None
        try:
            changes, params = data
            if not params:
                params = {}
            delta = self.delta(changes)
            if self.on_apply:
                result, error = self.on_apply(self, delta, params, privileges, user_info, enviroment)
            else:
                sql = delta.apply_sql(privileges)
                result, error = self.task.execute(sql)
        except Exception, e:
            error = e.message
            if not error:
                error = '%s: apply_changes error' % self.item_name
            print traceback.format_exc()
        if error:
            return {'error': error, 'result': None}
        else:
            return {'error': error, 'result': result}


    def field_by_id(self, id_value, field_name):
        return self.get_field_by_id((id_value, field_name))

    def get_field_by_id(self, params):
        id_value, field_name = params
        if isinstance(field_name, tuple) or isinstance(field_name, list):
            field_names = field_name
        else:
            field_names = [field_name]
        field_types = [self._field_by_name(field_name).data_type for field_name in field_names]
        fields_str = ', '.join(['"%s"."%s"' % (self.table_name.upper(), field_name.upper()) for field_name in field_names])
        sql = 'SELECT %s FROM "%s" WHERE ID = %s AND DELETED = 0' % (fields_str, self.table_name.upper(), id_value)
        rec = self.task.execute_select_one(sql)
        if rec:
            rec = list(rec)
            for i, val in enumerate(rec):
                if val is None:
                    if field_types[i] == common.TEXT:
                        rec[i] = ''
                    elif field_types[i] in [common.INTEGER, common.FLOAT, common.CURRENCY]:
                        rec[i] = 0
            if len(field_names) == 1:
                return rec[0]
            else:
                return rec

class ServerAbstractItem(object):
    def __init__(self):
        self.reports = []

    def init_reports(self):
        pass

    def get_reports_info(self):
        result = []
        for report in self.reports:
            result.append(report.ID)
        return result

    def register(self, func):
        setattr(self, func.__name__, func)


class ServerItem(Item, ServerAbstractItem, ServerDataset):
    def __init__(self, owner, name, caption, visible = True,
            table_name='', view_template='', edit_template='', filter_template='', soft_delete=True):
        Item.__init__(self, owner, name, caption, visible)
        ServerAbstractItem.__init__(self)
        ServerDataset.__init__(self, table_name, view_template, edit_template, filter_template, soft_delete)
        self.item_type_id = None

class ServerParam(DBField):
    def __init__(self, caption='', name='', data_type=common.INTEGER, item=None, lookup_field=None, required=True, visible=True, alignment=0):
        DBField.__init__(self)
        self.field_caption = caption
        self.param_caption = caption
        self.field_name = name
        self.param_name = name
        self.lookup_item = item
        self.lookup_field = lookup_field
        self.data_type = data_type
        if self.data_type == common.TEXT:
            self.field_size = 1000
        else:
            self.field_size = 0
        self.required = required
        self.alignment = alignment
        self.edit_visible = visible
        self.edit_value = None
        self.param_lookup_value = None

    def system_field(self):
        return False

    def get_data(self):
        return self.edit_value

    def set_data(self, value):
        self.edit_value = value

    def get_lookup_data(self):
        return self.param_lookup_value

    def set_lookup_data(self, value):
        self.param_lookup_value = value

    def do_before_changed(self, new_value, new_lookup_value):
        pass

    def do_on_change_lookup_field(self, lookup_value=None, slave_field_values=None):
        pass

    def raw_display_text(self):
        result = ''
        if self.lookup_item:
            result = self.lookup_text
        else:
            result = self.text
        return result

    def copy(self, owner):
        result = ServerParam(self.param_caption, self.field_name, self.data_type,
            self.lookup_item, self.lookup_field, self.required,
            self.edit_visible, self.alignment)
        result.set_info(self.get_info())
        result.owner = owner
        return result


class ServerReport(Report, ServerAbstractItem):
    def __init__(self, owner, name='', caption='', visible = True,
            table_name='', view_template='', edit_template='', filter_template=''):
        ServerAbstractItem.__init__(self)
        Report.__init__(self, owner, name, caption, visible)
        self.params = []
        self.template = view_template
        self.band_tags = []
        self.bands = {}
        self.header = None
        self.footer = None
        self.on_before_generate_report = None
        self.on_generate_report = None
        self.on_report_generated = None
        self.on_before_save_report = None

    def copy(self):
        result = self.__class__(self.owner, self.item_name, self.item_caption, self.visible,
            self.table_name, self.template, '', '');
        result.on_before_generate_report = self.on_before_generate_report
        result.on_generate_report = self.on_generate_report
        result.on_report_generated = self.on_report_generated
        result.on_before_save_report = self.on_before_save_report
        for param in self.params:
            new_param = param.copy(result)
            new_param.lookup_item = param.lookup_item
            result.params.append(new_param)
        for param in result.params:
            if param.master_field:
                param.master_field = result.get_master_field(params, param.master_field)
        return  result

    def add_param(self, caption='', name='', data_type=common.INTEGER, obj=None, obj_field=None, required=True, visible=True, value=None):
        self.params.append(ServerParam(caption, name, data_type, obj, obj_field, required, visible, value))

    def get_params_info(self):
        result = []
        for param in self.params:
            result.append(param.get_info())
        return result

    def get_edit_template(self):
        if self.edit_template:
            return self.edit_template
        elif self.owner:
            return self.owner.get_edit_template()

    def get_edit_ui(self):
        return common.ui_to_string(os.path.join(self.task.get_ui_path(), 'ui', self.get_edit_template()))

    def print_report(self, param_values, url, ext=None):
        copy_report = self.copy();
        return copy_report.generate(param_values, url, ext);

    def get_report_file_name(self, ext=None):
        if not ext:
            ext = 'ods'
        file_name = self.item_caption + '_' + datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S.%f') + '.' + ext
        file_name = escape(file_name, {':': '-', '/': '_', '\\': '_'})
        return os.path.abspath(os.path.join(self.task.work_dir, 'static', 'reports', file_name))

    def generate(self, param_values, url, ext):
        self.extension = ext
        self.url = url
        for i, param in enumerate(self.params):
            param.value = param_values[i]
        if self.on_before_generate_report:
            self.on_before_generate_report(self)
        if self.template:
            if not len(self.bands):
                self.parse_template()
            self.content_name = os.path.join(self.task.work_dir, 'reports', 'content%s.xml' % time.time())
            self.content = open(self.content_name, 'wb')
            try:
                #~ file_name = self.item_caption + '_' + datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S.%f') + '.ods'
                #~ file_name = escape(file_name, {':': '-', '/': '_', '\\': '_'})
                #~ self.report_filename = os.path.abspath(os.path.join(self.task.work_dir, 'static', 'reports', file_name))
                self.report_filename = self.get_report_file_name()
                file_name = os.path.basename(self.report_filename)
                static_dir = os.path.dirname(self.report_filename)
                if not os.path.exists(static_dir):
                    os.makedirs(static_dir)
                if self.header:
                    self.content.write(self.header)
                if self.on_generate_report:
                    self.on_generate_report(self)
                if self.footer:
                    self.content.write(self.footer)
                self.save()
            finally:
                if not self.content.closed:
                    self.content.close()
                os.remove(self.content_name)
            if ext and (ext != 'ods'):
                converted = False
                if self.owner.on_convert_report:
                    try:
                        self.owner.on_convert_report(self)
                        converted = True
                    except:
                        pass
                if not converted:
                    # OpenOffice must be running in server mode
                    # soffice --headless --accept="socket,host=127.0.0.1,port=2002;urp;"
                    ext_file_name = self.report_filename.replace('.ods', '.' + ext)
                    try:
                        from third_party.DocumentConverter import DocumentConverter
                        converter = DocumentConverter()
                        converter.convert(self.report_filename, ext_file_name)
                        converted = True
                    except:
                        pass
                if not converted:
                    try:
                        from subprocess import Popen, STDOUT, PIPE
                        convertion = Popen(['soffice', '--headless', '--convert-to', ext,
                            self.report_filename, '--outdir', os.path.join(self.task.work_dir, 'static', 'reports') ],
                            stderr=STDOUT,stdout = PIPE)#, shell=True)
                        out, err = convertion.communicate()
                        converted = True
                    except:
                        pass
                converted_file = self.report_filename.replace('.ods', '.' + ext)
                if converted and os.path.exists(converted_file):
                    self.delete_report(self.report_filename)
                    file_name = file_name.replace('.ods', '.' + ext)
            self.report_filename = os.path.join(self.task.work_dir, 'static', 'reports', file_name)
            self.report_url = self.report_filename
            if self.url:
                self.report_url = '%s/static/reports/%s' % (self.url, file_name)
        else:
            if self.on_generate_report:
                self.on_generate_report(self)
        if self.on_report_generated:
            self.on_report_generated(self)
        return self.report_url

    def delete_report(self, file_name):
        report_name = os.path.join(self.task.work_dir, 'static', 'reports', file_name)
        os.remove(report_name)

    def parse_template(self):
        self.template_name = os.path.join(self.task.work_dir, 'reports', self.template)
        z = zipfile.ZipFile(self.template_name, 'r')
        try:
            data = unicode(z.read('content.xml'), 'utf-8')
        finally:
            z.close()

        self.band_tags = []
        self.bands = {}
        repeated_rows = None
        if data:
            dom = parseString(data)
            try:
                tables = dom.getElementsByTagName('table:table')
                if len(tables) > 0:
                    table = tables[0]
                    for child in table.childNodes:
                        if child.nodeName == 'table:table-row':
                            repeated = child.getAttribute('table:number-rows-repeated')
                            if repeated and repeated.isdigit():
                                repeated_rows = repeated
                            for row_child in child.childNodes:
                                if row_child.nodeName == 'table:table-cell':
                                    text = row_child.getElementsByTagName('text:p')
                                    if text.length > 0:
                                        self.band_tags.append(text[0].childNodes[0].nodeValue)
                                    break

                assert len(self.band_tags) > 0, u'No bands in report template'
                positions = []
                start = 0
                for tag in self.band_tags:
                    text = str('>%s<' % tag)
                    i = data.find(text)
                    i = data.rfind('<table:table-row', start, i)
                    positions.append(i)
                    start = i
                if repeated_rows and int(repeated_rows) > 1000:
                    i = data.find(repeated_rows)
                    i = data.rfind('<table:table-row', start, i)
                    self.band_tags.append('$$$end_of_report')
                    positions.append(i)
                self.header = data[0:positions[0]]
                for i, tag in enumerate(self.band_tags):
                    start = positions[i]
                    try:
                        end = positions[i + 1]
                    except:
                        end = data.find('</table:table>', start)
                    self.bands[tag] = data[start: end].replace(str(tag), '')
                self.footer = data[end:len(data)]
            finally:
                dom.unlink()
                del(dom)

    def print_band(self, band, dic=None, update_band_text=None):
        text = self.bands[band]
        if dic:
            d = dic.copy()
            for key, value in d.items():
                if type(value) in (str, unicode):
                    d[key] = escape(value)
            try:
                cell_start = 0
                cell_start_tag = '<table:table-cell'
                cell_type_tag = 'office:value-type="string"'
                calcext_type_tag = 'calcext:value-type="string"'
                start_tag = '<text:p>'
                end_tag = '</text:p>'
                while True:
                    cell_start = text.find(cell_start_tag, cell_start)
                    if cell_start == -1:
                        break
                    else:
                        start = text.find(start_tag, cell_start)
                        if start != -1:
                            end = text.find(end_tag, start + len(start_tag))
                            if end != -1:
                                text_start = start+len(start_tag)
                                text_end = end
                                cell_text = text[text_start:text_end]
                                cell_text_start = cell_text.find('%(', 0)
                                if cell_text_start != -1:
                                    end = cell_text.find(')s', cell_text_start + 2)
                                    if end != -1:
                                        end += 2
                                        val = cell_text[cell_text_start:end]
                                        key = val[2:-2]
                                        value = d.get(key)
                                        if not value is None:
                                            val = val % d
                                            if type(value) == float:
                                                val = val.replace('.', common.DECIMAL_POINT)
                                        else:
                                            if not key in d.keys():
                                                print 'Report: "%s" band: "%s" key "%s" not found in the dictionary' % (self.item_name, band, key)
                                        cell_text = cell_text[:cell_text_start] + val + cell_text[end:]
                                        text = text[:text_start] + cell_text + text[text_end:]
                                        if type(value) in (int, float):
                                            start_text = text[cell_start:start]
                                            office_value = str(value)
                                            start_text = start_text.replace(cell_type_tag, 'office:value-type="float" office:value="%s"' % office_value)
                                            start_text = start_text.replace(calcext_type_tag, 'calcext:value-type="float"')
                                            text = text[:cell_start] + start_text + text[start:]
                        cell_start += 1
                if update_band_text:
                    text = update_band_text(text)
            except Exception, e:
                print traceback.format_exc()
                print ('Report: "%s" band: "%s" error: "%s"') % (self.item_name, band, e)
        self.content.write(text.encode('utf-8'))

    def save(self):
        self.content.close()
        z = None
        self.zip_file = None
        try:
            self.zip_file = zipfile.ZipFile(self.report_filename, 'w', zipfile.ZIP_DEFLATED)
            z = zipfile.ZipFile(self.template_name, 'r')
            if self.on_before_save_report:
                self.on_before_save_report(self)
            for file_name in z.namelist():
                data = z.read(file_name)
                if file_name == 'content.xml':
                    self.zip_file.write(self.content_name, file_name)
                else:
                    self.zip_file.writestr(file_name, data)
        finally:
            if z:
                z.close()
            if self.zip_file:
                self.zip_file.close()
#            os.remove(self.content_name)

    def cur_to_str(self, value):
        return common.cur_to_str(value)

    def date_to_str(self, value):
        return common.date_to_str(value)

    def datetime_to_str(self, value):
        return common.datetime_to_str(value)



delta_result = None

def execute_sql(db_type, db_database, db_user, db_password,
    db_host, db_port, db_encoding, connection, command,
    params=None, result_set=None, call_proc=False, commit=True):

    def execute_command(cursor, command, params=None):
        try:
            #~ print ''
            #~ print command, params

            result = None
            if params:
                cursor.execute(command, params)
            else:
                cursor.execute(command)
            if result_set == 'ONE':
                result = cursor.fetchone()
            elif result_set == 'ALL':
                result = cursor.fetchall()
            return result
        except Exception, x:
            print '\nError: %s\n command: %s\n params: %s' % (str(x), command, params)
            raise

    def get_next_id(cursor, sql):
        cursor.execute(sql)
        rec = cursor.fetchone()
        return int(rec[0])

    def execute_delta(cursor, command):

        def process_delta(delta, master_rec_id, result):
            ID, sqls = delta
            result['ID'] = ID
            changes = []
            result['changes'] = changes
            for sql in sqls:
                (command, params, info), details = sql
                if info:
                    rec_id = info['id']
                    if rec_id:
                        if info['change_id_sql'] and info['next_id_sql']:
                            next_id = get_next_id(cursor, info['next_id_sql'])
                            if next_id < rec_id:
                                cursor.execute(info['change_id_sql'])
                    else:
                        if info['next_id_sql']:
                            rec_id = get_next_id(cursor, info['next_id_sql'])
                            params[info['id_index']] = rec_id
                    if info['status'] == common.RECORD_INSERTED and info['owner_rec_id_index']:
                        params[info['owner_rec_id_index']] = master_rec_id
                    if command:
                        execute_command(cursor, command, params)
                    if db_type == common.SQLITE and not info['status'] == common.RECORD_DELETED and not rec_id:
                        rec_id = cursor.lastrowid
                    result_details = []
                    if rec_id:
                        changes.append({'log_id': info['log_id'], 'rec_id': rec_id, 'details': result_details})
                    for detail in details:
                        result_detail = {}
                        result_details.append(result_detail)
                        process_delta(detail, rec_id, result_detail)
                else:
                    if command:
                        execute_command(cursor, command, params)

        global delta_result
        delta = command['delta']
        delta_result = {}
        process_delta(delta, None, delta_result)

    def execute_list(cursor, command):
        res = None
        if command:
            for com in command:
                if com:
                    if isinstance(com, unicode) or isinstance(com, str):
                        res = execute_command(cursor, com)
                    elif isinstance(com, list):
                        res = execute_list(cursor, com)
                    elif isinstance(com, dict):
                        res = execute_delta(cursor, com)
                    elif isinstance(com, tuple):
                        res = execute_command(cursor, com[0], com[1])
                    else:
                        raise Exception, 'server_classes execute_list: invalid argument - command: %s' % command
            return res

    def execute(connection):
        global delta_result
        result = None
        error = None
        try:
            cursor = connection.cursor()
            if call_proc:
                try:
                    cursor.callproc(command, params)
                    result = cursor.fetchone()
                except Exception, x:
                    print '\nError: %s in command: %s' % (str(x), command)
                    raise
            else:
                if isinstance(command, str) or isinstance(command, unicode):
                    result = execute_command(cursor, command, params)
                elif isinstance(command, dict):
                    res = execute_delta(cursor, command)
                elif isinstance(command, list):
                    result = execute_list(cursor, command)
                elif isinstance(command, tuple):
                    result = execute_command(cursor, command[0], command[1])
            if commit:
                connection.commit()
            if delta_result:
                result = delta_result
        except Exception, x:
            try:
                if connection:
                    connection.rollback()
                    connection.close()
                error = str(x)
                if not error:
                    error = 'Execute error'
                print traceback.format_exc()
            finally:
                connection = None
        return connection, (result, error)

    global delta_result
    delta_result = None
    if not db_host:
        db_host = 'localhost'
    if db_type == common.POSTGRESQL:
        if connection is None:
            import psycopg2
            connection = psycopg2.connect(database=db_database, user=db_user, password=db_password, host=db_host, port=db_port)
        return execute(connection)
    elif db_type == common.MYSQL:
        if connection is None:
            import MySQLdb
            connection = MySQLdb.connect(host=db_host, user=db_user, passwd=db_password, db=db_database)
            cursor = connection.cursor()
            cursor.execute("SET SESSION SQL_MODE=ANSI_QUOTES;")
        return execute(connection)
    elif db_type == common.FIREBIRD:
        if connection is None:
            import fdb
            connection = fdb.connect(host=db_host, database=db_database, user=db_user, password=db_password, charset=db_encoding)
        return execute(connection)
    elif db_type == common.SQLITE:
        if connection is None:
            import sqlite3
            connection = sqlite3.connect(db_database)
            cursor = connection.cursor()
            cursor.execute("PRAGMA foreign_keys = ON")
        return execute(connection)

def process_request(name, queue, db_type, db_database, db_user, db_password, db_host, db_port, db_encoding, mod_count):
    con = None
    counter = 0
    while True:
        request = queue.get()
        if request:
#            print name, 'process id:', os.getpid()
            result_queue = request['queue']
            command = request['command']
            params = request['params']
            result_set = request['result_set']
            call_proc = request['call_proc']
            commit = request['commit']
            cur_mod_count = request['mod_count']
            if cur_mod_count != mod_count or counter > 1000:
                if con:
                    con.rollback()
                    con.close()
                con = None
                mod_count = cur_mod_count
                counter = 0
            if command == 'QUIT':
                if con:
                    con.commit()
                    con.close()
                result_queue.put('QUIT')
                break
            else:
                con, result = execute_sql(db_type, db_database, db_user, db_password,
                    db_host, db_port, db_encoding, con, command, params, result_set, call_proc, commit)
                counter += 1
                result_queue.put(result)

class AbstractServerTask(Task, ServerAbstractItem):
    def __init__(self, name, caption, template, edit_template, db_type, db_database = '',
            db_user = '', db_password = '', host='', port='', encoding='', con_pool_size=1):
        Task.__init__(self, None, None, None, None)
        self.items = []
        self.ID = None
        self.item_name = name
        self.item_caption = caption
        self.template = template
        self.db_type = db_type
        self.db_database = db_database
        self.db_user = db_user
        self.db_password = db_password
        self.db_host = host
        self.db_port = port
        self.db_encoding = encoding
        self.work_dir = os.getcwd()
        self.con_pool_size = 0
        self.mod_count = 0
#        self.modules = []
        self.processes = []
        self._busy = 0
        if sys.db_multiprocessing and con_pool_size:
            self.queue = multiprocessing.Queue()
            self.manager = multiprocessing.Manager()
            self.con_pool_size = con_pool_size
            self.create_connection_pool()
        else:
            self.connection = None

    def create_connection_pool(self):
        if self.con_pool_size:
            for i in range(self.con_pool_size):
                p = multiprocessing.Process(target=process_request, args=(self.item_name,
                    self.queue, self.db_type, self.db_database, self.db_user,
                    self.db_password, self.db_host, self.db_port,
                    self.db_encoding, self.mod_count))
                self.processes.append(p)
                p.daemon = True
                p.start()

    def release_connection_pool(self):
        if self.con_pool_size:
            for i in range(self.con_pool_size):
                self.execute('QUIT')
            self.processes = []

    def execute(self, command, params=None, result_set=None, call_proc=False, commit=True):
        if self.con_pool_size:
            result_queue = self.manager.Queue()
            request = {}
            request['queue'] = result_queue
            request['command'] = command
            request['command'] = command
            request['params'] = params
            request['result_set'] = result_set
            request['call_proc'] = call_proc
            request['commit'] = commit
            request['mod_count'] = self.mod_count
            self._busy += 1
            try:
                self.queue.put(request)
                result = result_queue.get()
            finally:
                self._busy -= 1
            return result
        else:
            result = execute_sql(self.db_type, self.db_database, self.db_user,
                self.db_password, self.db_host, self.db_port,
                self.db_encoding, self.connection, command, params, result_set, call_proc, commit)
            self.connection = result[0]
            return result[1]

    def callproc(self, command, params=None):
        result_set, error = self.execute(command, params, call_proc=True)
        if not error:
            return result_set

    def execute_select(self, command, params=None):
        result, error = self.execute(command, params, result_set='ALL', commit=False)
        if error:
            raise Exception, error
        else:
            return result


    def execute_select_one(self, command, params=None):
        result, error = self.execute(command, params, result_set='ONE', commit=False)
        if error:
            raise Exception, error
        else:
            return result

    def get_module_name(self):
        return str(self.item_name + '_' + 'server')

    def compile_item(self, item):
        item.module_name = None
        code = item.server_code
        item.module_name = item.get_module_name()
        item_module = type(sys)(item.module_name)
        item_module.__dict__['this'] = item
        sys.modules[item.module_name] = item_module
        if item.owner:
            sys.modules[item.owner.get_module_name()].__dict__[item.module_name] = item_module
        if code:
            try:
                code = code.encode('utf-8')
            except:
                pass
            try:
                comp_code = compile(code, item.module_name, "exec")
            except Exception, e:
                print e
            exec comp_code in item_module.__dict__
            funcs = inspect.getmembers(item_module, inspect.isfunction)
            item._events = []
            for func_name, func in funcs:
                item._events.append((func_name, func))
                setattr(item, func_name, func)
        del code

    def login(self, params):
        return 1

    def __getattr__(self, name):
        if self.item_by_name(name):
            return self.item_by_name(name)
        else:
            raise AttributeError (self.item_name + ' AttributeError: ' + name)

    def add_item(self, item):
        self.items.append(item)
        item.owner = self
        return item

    def find_item(self, g_index, i_index):
        return self.items[g_index].items[i_index]

    def get_ui_file(self, file_name):
        return common.ui_to_string(os.path.join(self.get_ui_path(), 'ui', file_name))

    def get_ui(self):
        if self.template:
            return common.ui_to_string(os.path.join(self.get_ui_path(), 'ui', self.template))

    def copy_database_data(self, db_type, db_database=None, db_user=None, db_password=None,
        db_host=None, db_port=None, db_encoding=None):
        connection = None
        limit = 1024
        for group in self.items:
            for item in group.items:
                if item.item_type != 'report':
                    self.execute('DELETE FROM %s' % item.table_name)
                    item.open(expanded=False, open_empty=True)
                    params = {'__fields': [], '__filters': [], '__expanded': False, '__loaded': 0, '__limit': 0}
                    sql = item.get_record_count_query(params, db_type)
                    connection, (result, error) = \
                    execute_sql(db_type, db_database, db_user, db_password,
                        db_host, db_port, db_encoding, connection, sql, None, 'ALL')
                    record_count = result[0][0]
                    loaded = 0
                    while True:
                        params['__loaded'] = loaded
                        params['__limit'] = limit
                        sql = item.get_select_statement(params, db_type)
                        connection, (result, error) = \
                        execute_sql(db_type, db_database, db_user, db_password,
                            db_host, db_port, db_encoding, connection, sql, None, 'ALL')
                        if not error:
                            for i, r in enumerate(result):
                                item.append()
                                j = 0
                                for field in item.fields:
                                    if not field.master_field:
                                        field.value = r[j]
                                        j += 1
                                item.post()
                            item.apply()
                        else:
                            raise Exception, error
                        loaded = len(result)
                        loaded += loaded
                        print 'coping table %s: %d%%' % (item.item_name, int(loaded * 100 / record_count))
                        if loaded == 0 or loaded < limit:
                            break


class ServerTask(AbstractServerTask):
    def __init__(self, name, caption, template, edit_template,
        db_type, db_database = '', db_user = '', db_password = '',
        host='', port='', encoding='', con_pool_size=4):
        AbstractServerTask.__init__(self, name, caption, template, edit_template,
            db_type, db_database, db_user, db_password,
            host, port, encoding, con_pool_size)
        self.on_created = None
        self.on_login = None
        self.on_get_user_info = None
        self.on_logout = None
        self.on_ext_request = None

    def get_ui_path(self):
        filepath = os.getcwd()
        return filepath.decode('utf-8')

    def find_user(self, login, password_hash=None):
        return self.admin.find_user(login, password_hash);

class AdminServerTask(AbstractServerTask):
    def __init__(self, name, caption, template, edit_template,
        db_type, db_database = '', db_user = '', db_password = '',
        host='', port='', encoding=''):
        AbstractServerTask.__init__(self, name, caption, template, edit_template,
            db_type, db_database, db_user, db_password, host, port, encoding, 1)
        filepath, filename = os.path.split(__file__)
        self.cur_path = filepath

    def get_ui_path(self):
        return self.cur_path.decode('utf-8')


class ServerGroup(Group, ServerAbstractItem):
    def __init__(self, owner, name, caption, view_template = None, edit_template = None, filter_template = None, visible = True, item_type_id=0):
        ServerAbstractItem.__init__(self)
        Group.__init__(self, owner, name, caption, True, item_type_id)
        self.ID = None
        self.view_template = view_template
        self.edit_template = edit_template
        self.filter_template = filter_template
        if item_type_id == common.REPORTS_TYPE:
            self.on_convert_report = None

    def __getattr__(self, name):
        if self.find(name):
            return self.find(name)
        else:
            raise AttributeError (self.item_name + ' AttributeError: ' + name)

    def get_view_template(self):
        return self.view_template

    def get_edit_template(self):
        return self.edit_template

    def get_filter_template(self):
        return self.filter_template

    def add_ref(self, name, caption, table_name, visible = True, view_template = '', edit_template = '', filter_template='', soft_delete=True):
        result = ServerItem(self, name, caption, visible, table_name, view_template, edit_template, filter_template, soft_delete)
        result.item_type_id = common.CATALOG_TYPE
        return result

    def add_journal(self, name, caption, table_name, visible = True, view_template = '', edit_template = '', filter_template='', soft_delete=True):
        result = ServerItem(self, name, caption, visible, table_name, view_template, edit_template, filter_template, soft_delete)
        result.item_type_id = common.JOURNAL_TYPE
        return result

    def add_table(self, name, caption, table_name, visible = True, view_template = '', edit_template = '', filter_template='', soft_delete=True):
        result = ServerItem(self, name, caption, visible, table_name, view_template, edit_template, filter_template, soft_delete)
        result.item_type_id = common.TABLE_TYPE
        return result

    def add_report(self, name, caption, table_name, visible = True, view_template = '', edit_template = '', filter_template='', soft_delete=True):
        result = ServerReport(self, name, caption, visible, table_name, view_template, edit_template, filter_template)
        result.item_type_id = common.REPORT_TYPE
        return result


class ServerDetail(ServerAbstractItem, Detail, ServerDataset):
    def __init__(self, owner, name, caption, table_name):
        ServerAbstractItem.__init__(self)
        Detail.__init__(self, owner, name, caption, True)
        ServerDataset.__init__(self, table_name)
        self.prototype = self.task.item_by_name(self.item_name)
        self.master = owner

    def init_fields(self):
        for field in self.prototype._fields:
            self._fields.append(field.copy(self))

    def get_gen_name(self, db_type):
        return self.prototype.get_gen_name(db_type)

    def do_internal_post(self):
        return {'success': True, 'id': None, 'message': '', 'detail_ids': None}

    def where_clause(self, query, db_type):
        owner_id = query['__owner_id']
        owner_rec_id = query['__owner_rec_id']
        if type(owner_id) == int and type(owner_rec_id) == int:
            result = super(ServerDetail, self).where_clause(query, db_type)
            clause = '"%s"."OWNER_ID"=%s AND "%s"."OWNER_REC_ID"=%s' % \
            (self.table_name.upper(), str(owner_id), self.table_name.upper(), str(owner_rec_id))
            if result:
                result += ' AND ' + clause
            else:
                result = ' WHERE ' + clause
            return self.set_case(db_type, result)
        else:
            raise Exception, 'Invalid request parameter'

    def get_filters(self):
        return self.prototype.filters

    #~ def get_reports_info(self):
        #~ return []



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys, os
sys.db_multiprocessing = True

import cPickle
import json
import traceback
try:
    import SimpleHTTPServer
except:
    import SimpleHTTPServer
try:
    import web
except:
    import jam.third_party.web as web

import common

urls = (
    '/api(.*)', 'Api',
    '/ext(.*)', 'Ext',
    '/upload(.*)', 'Upload',
    '/(.*)', 'Index',
)

web.config.debug = True
app = web.application(urls, globals(), autoreload=False)

class Request(object):
    def __init__(self):
        self.web_request = None

    def get_request(self):
        data = web.data()
        self.web_request = data[0] == '1'
        if self.web_request:
            result = json.loads(data[1:])
        else:
            result = cPickle.loads(data[1:].decode('zlib'))
        return result

    def prepare_response(self, r):
        if self.web_request:
            accepts_gzip = 0
            try:
                if web.ctx.env.get("HTTP_ACCEPT_ENCODING").find("gzip") != -1:
                    accepts_gzip = 1
            except:
                pass
            buff = json.dumps(r, default=common.json_defaul_handler)
            web.header('Content-Type', 'application/json')
            if accepts_gzip:
                buff = common.compressBuf(buff)
                web.header('Content-encoding', 'gzip')
                web.header('Content-Length', str(len(buff)))
                return buff
            else:
                return buff
        else:
            web.header('Content-encoding', 'deflate')
            return cPickle.dumps(r, 2).encode('zlib')


class Api(Request):
    def __init__(self):
        Request.__init__(self)

    def POST(self, name):
        r = {'result': None, 'error': None}
        q = self.get_request()
        try:
            method = q['method']
            if method == 'send_request':
                request = q['params'][0]
                user_id = q['params'][1]
                task_id = q['params'][2]
                item_id = q['params'][3]
                params = q['params'][4]
                r['result'] = web.server.process_request(web.ctx.env, request, user_id, task_id, item_id, params)
                if task_id == 0:
                    if request == 'exit' and r['result']:
                        sys.exit(0)
            else:
                r['result'] = None
        except Exception, e:
            print traceback.format_exc()
            r['error'] = e.message
        return self.prepare_response(r)

class Ext(Request):
    def __init__(self):
        Request.__init__(self)

    def get_request(self):
        data = web.data()
        self.web_request = True
        result = json.loads(data)
        return result

    def POST(self, name):
        r = {'result': None, 'error': None}
        q = self.get_request()
        try:
            request = name
            user_id = None
            task_id = None
            item_id = None
            params = q
            ext = True
            r['result'] = web.server.process_request(web.ctx.env, request, user_id, task_id, item_id, params, ext)
        except Exception, e:
            print traceback.format_exc()
            r['error'] = e.message
        return self.prepare_response(r)

class Upload:
    def POST(self, name):
        data = web.data()
        header = []
        header_str = ''
        length = 0
        string = ''
        for s in data:
            header_str += s
            if s == ';':
                if len(header) == 0:
                    length = int(string)
                header.append(int(string))
                if len(header) == 2 * (length + 1):
                    break;
                string = ''
            else:
                string += s
        start = len(header_str)
        path = os.path.join(os.getcwd(), os.path.normpath(data[start: start + header[1]]))
        if not os.path.exists(path):
            os.makedirs(path)
        start = start + header[1]
        for i in range(length):
            index = 2 * i + 2
            file_name = data[start: start + header[index]]
            start = start + header[index]
            index += 1
            content = data[start: start + header[index]]
            file_name = os.path.join(path, file_name)
            with open(file_name, 'wb') as f:
                f.write(content)
            start = start + header[index]

class Index:
    def GET(self, name):
        if not name:
            name = 'index.html'
        file_name = os.path.normpath(name)
        if file_name in ['index.html', 'admin.html']:
            web.server.check_task_client_modified(file_name)
        file_name = web.server.check_file_name(file_name)
        base, ext = os.path.splitext(file_name)
        content_type = common.mime_type_by_ext(ext)
        if file_name and os.path.isfile(file_name):
            with open(os.path.join(file_name), 'rb') as f:
                web.header('Content-Type', content_type)
#                web.header('Cache-Control', 'no-cache')
                return f.read()
        else:
            return ''

def run(server):
    me = common.SingleInstance()
    web.server = server
    server.app = app
    app.run()

